/*
 * Copyright 2006-2007, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <math.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>


extern int Localmode;
int Port = 7536;
char *McastAddr = "224.0.0.75";

int 
setup_source(void)
{
	if (Localmode) {
		FILE *in;
		char execname[PATH_MAX], bin[PATH_MAX];
		int fid, err;

		strcpy(execname,getexecname());
		strcpy(bin,dirname(execname));
		strcat(bin,"/../sbin/sysstatd -l");
		dbug("attaching to %s\n",bin);
		in = popen(bin,"r");
		if (in == 0)
			error("could not attach sysstatd locally: %s\n",strerror(errno));
		fid = fileno(in);
		err = fcntl(fid,F_SETFL,O_NONBLOCK);
		if (err == 0)
			warn("could not set pipe to nonblocking: %s\n",strerror(errno));
		dbug("successfully attached to sysstatd via stream %p (fileno %d)\n",in,fid);
		return fid;
	} else {
		struct sockaddr_in MA;
		struct ip_mreq mreq;
		int on = 1, fl;
		int sock;

		sock = socket(AF_INET, SOCK_DGRAM, 0);
		if (-1 == sock)
			error("cannot create socket: %s\n",strerror(errno));
		fl = fcntl(sock,F_GETFL,0);
		fcntl(sock,F_SETFL,fl | O_NONBLOCK);
		if (-1 == setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)))
			error("cannot set socket to reuse address: %s\n",strerror(errno));
		bzero(&MA,sizeof(MA));
		MA.sin_family = AF_INET;
		MA.sin_port = htons(Port);
		if (-1 == inet_pton(AF_INET, McastAddr, &MA.sin_addr))
			error("cannot set multicast address: %s\n",strerror(errno));
		if (-1 == bind(sock,(struct sockaddr *)&MA,sizeof(MA)))
			error("cannot bind socket: %s\n",strerror(errno));

		memcpy(&mreq.imr_multiaddr.s_addr,(void *)&MA.sin_addr,sizeof(struct in_addr));
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
		if (-1 == setsockopt(sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)))
			error("could not join multicast group: %s\n",strerror(errno));
		return sock;
	}
}


static unsigned long long
get_unsigned(const char *restrict key, const char *restrict value)
{
	unsigned long long ret = 0;
	char c;
	assert(value);
	c = *value;
	if ((c < '0') || (c > '9')) {
		warn("expected an unsigned integer, got: %s\t%s\n",key,value);
		return 0;
	}
	do {
		c -= '0';
		++value;
		ret *= 10;
		ret += c;
		c = *value;
	} while ((c >= '0') && (c <= '9'));
	return ret;
}


static void
parse_dev(char *d, sysstat_t *s)
{
	char *dev = d;
	fsstat_t *l;
	fsstat_t *fs = calloc(1,sizeof(fsstat_t));
	if (s->fs) {
		l = s->fs;
		while (l->next) l = l->next;
		l->next = fs;
	} else {
		s->fs = fs;
	}
	d = strchr(d,'\t');
	assert(d);
	*d = 0;
	++d;
	fs->dev = strdup(dev);
	fs->ops_rate = -1;
	//dbug("parse_dev(%s,%s,%p)\n",dev,m,s);
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		//dbug("parse_dev(): key = \"%s\", value = \"%s\"\n",key,value);
		if (0 == strcmp("mnt",key)) {
			fs->mnt = strdup(value);
		} else if (0 == strcmp("fs",key)) {
			if (strlen(value) > 7) 
				warn("protocol error: %s\t%s\n",key,value);
			else
				strcpy(fs->typ,value);
		} else if (0 == strcmp("fr",key)) {
			fs->free = get_unsigned(key,value); 
		} else if (0 == strcmp("bl",key)) {
			fs->blocks = get_unsigned(key,value); 
		} else if (0 == strcmp("ops",key)) {
			fs->ops_rate = strtod(value,0); 
		} else if (0 == strcmp("inr",key)) {
			fs->read_rate = strtod(value,0); 
		} else if (0 == strcmp("outr",key)) {
			fs->write_rate = strtod(value,0); 
		} else
			warn("unkown fs subfield: %s\t%s\n",key,value);
	}
	if (fs->mnt == 0)
		fs->mnt = strdup("");
}


static void
parse_user(char *d, sysstat_t *s)
{
	userstat_t *l = 0;
	userstat_t *us = calloc(1,sizeof(userstat_t));
	if (1 != sscanf(d,"%lu",&us->uid)) {
		warn("protocol error: uid\t%s\n",d);
		free(us);
		return;
	}
	if (s->us) {
		l = s->us;
		while (l->next) l = l->next;
		l->next = us;
	} else {
		s->us = us;
	}
	d = strchr(d,'\t');
	assert(d);
	++d;
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		//dbug("parse_user(): key = \"%s\", value = \"%s\"\n",key,value);
		if (*(key + 1) == 0) {
			switch (*key) {
			case 'N':
				us->nproc = get_unsigned(key,value); 
				break;
			case 'L':
				us->nlwp = get_unsigned(key,value); 
				break;
			case 'O':
				us->oncpu = get_unsigned(key,value); 
				break;
			case 'R':
				us->runnable = get_unsigned(key,value); 
				break;
			case 'S':
				us->sleeping = get_unsigned(key,value); 
				break;
			case 'T':
				us->stopped = get_unsigned(key,value); 
				break;
			case 'Z':
				us->zombie = get_unsigned(key,value); 
				break;
			default:
				warn("unknown single char key %c\n",*key);
			}
		} else if (0 == strcmp("user",key)) {
			us->username = strdup(value);
		} else if (0 == strcmp("sec",key)) {
			if (sscanf(value,"%ld",&us->time) != 1) 
				warn("sec field error: %s\t%s\n",key,value);
		} else if (0 == strcmp("RSS",key)) {
			us->rss = get_unsigned(key,value); 
		} else if (0 == strcmp("SIZE",key)) {
			us->size = get_unsigned(key,value); 
		} else if (0 == strcmp("HS",key)) {
			us->h_s = get_unsigned(key,value); 
		} else if (0 == strcmp("%C",key)) {
			us->pctcpu = strtod(value,0); 
		} else if (0 == strcmp("%M",key)) {
			us->pctmem = strtod(value,0); 
		} else
			warn("unkown user subfield: %s\t%s\n",key,value);
	}
	if (us->username == 0)
		us->username = strdup("<unknown>");
}


static void
parse_cpurates(char *d, sysstat_t *s)
{
	int id;
	cpurates_t *r;

	id = get_unsigned(0,d);
	r = s->cpurates;
	while (r && (r->id != id)) {
		r = r->next;
	}
	if (r == 0) {
		r = calloc(1,sizeof(cpurates_t));;
		r->next = s->cpurates;
		s->cpurates = r;
	}
	r->id = id;
	d = strchr(d,'\t');
	assert(d);
	*d = 0;
	++d;
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		if (0 == strcmp("usr",key))
			r->user = strtod(value,0);
		else if (0 == strcmp("sys",key))
			r->kernel = strtod(value,0);
		else if (0 == strcmp("idl",key))
			r->idle = strtod(value,0);
		else if (0 == strcmp("intr",key))
			r->intr = strtod(value,0);
		else if (0 == strcmp("ithr",key))
			r->intrthread = strtod(value,0);
		else if (0 == strcmp("iblk",key))
			r->intrblk = strtod(value,0);
		else if (0 == strcmp("xcal",key))
			r->xcalls = strtod(value,0);
		else if (0 == strcmp("migr",key))
			r->migrate = strtod(value,0);
		else if (0 == strcmp("csw",key))
			r->pswitch = strtod(value,0);
		else if (0 == strcmp("icsw",key))
			r->inv_swtch = strtod(value,0);
		else if (0 == strcmp("smtx",key))
			r->mxae = strtod(value,0);
		else if (0 == strcmp("rwrf",key))
			r->rwrf = strtod(value,0);
		else if (0 == strcmp("rwwf",key))
			r->rwwf = strtod(value,0);
		else if (0 == strcmp("asf",key))
			r->as_fault = strtod(value,0);
		else if (0 == strcmp("hatf",key))
			r->hat_fault = strtod(value,0);
		else if (0 == strcmp("majf",key))
			r->majfault = strtod(value,0);
		else if (0 == strcmp("sysc",key))
			r->syscall = strtod(value,0);
		else
			warn("unknown cpu rates subfield \"%s\"\n",key);
	}
	dbug("received cpurates for cpu %d\n",r->id);
}


static char *get_pair(char *str, char **key, char **value)
{
	if ((str == 0) || (*str == 0))
		return 0;
	*key = str;
	str = strchr(str,'\t');
	if (str == 0) {
		warn("key without value (%s)\n",*key);
		return 0;
	}
	*str = 0;
	++str;
	*value = str;
	str = strchr(*value,'\n');
	if (str == 0) {
		warn("unterminated key %s\n",*key);
		return 0;
	}
	*str = 0;
	return ++str;
}


sysstat_t *
receive_stats(int sock)
{
	static char msg[4172+1], *m;	/* maximum size of one packet plus \0 */
	/* make key and value big enough, so that buffer overflow attacks have
	 * no chance */
	char *key,*value;
	int num;
	struct sockaddr_in from;
	socklen_t len;
	sysstat_t *s;

	if (Localmode)
		num = read(sock,&msg,sizeof(msg));
	else
		num = recvfrom(sock,&msg,sizeof(msg),0,(struct sockaddr *)&from,&len);
	dbug("received %d bytes\n",num);
	if ((num == -1) && (errno == EWOULDBLOCK))
		return 0;
	if (num == -1)
		error("cannot read from socket: %s\n",strerror(errno));
	msg[num] = '\0';
	dbug("packet:\n%s\n",msg);
	m = get_pair(msg,&key,&value);
	if (key == 0)
		error("cannot attach to sysstatd\n");
	if (strcmp(key,"sysstat") || strcmp(value,"1")) {
		warn("invalid protcol\n");
		return 0;
	}
	s = calloc(1,sizeof(sysstat_t));
	assert(s);
	s->IpInR = s->IpOutR = -1;
	while (m = get_pair(m,&key,&value)) {
		//dbug("key: \"%s\", value: \"%s\"\n",key,value);
		switch (*key) {
		case 'a':
			if (strcmp("anonfree",key) == 0)
				s->AnonfreeR = strtod(value,0);
			else if (strcmp("anonpgin",key) == 0)
				s->AnonpginR = strtod(value,0);
			else if (strcmp("anonpgout",key) == 0)
				s->AnonpgoutR = strtod(value,0);
			else
				goto unknown;
			continue;

		case 'c':
			if (strcmp("cpu",key) == 0)
				parse_cpurates(value,s);
			else if (strcmp("cpumigrate",key) == 0)
				s->MigrateR = strtod(value,0);
			else if (strcmp("call",key) == 0)
				s->SyscallR = strtod(value,0);
			else
				goto unknown;
			continue;

		case 'd':
			if (strcmp("dev",key) == 0) {
				parse_dev(value,s);
				++s->numfs;
			} else
				goto unknown;
			continue;

		case 'D':
			if (strcmp("DiskI",key) == 0)
				s->DiskIR = strtod(value,0);
			else if (strcmp("DiskO",key) == 0)
				s->DiskOR = strtod(value,0);
			else if (strcmp("Disk",key) == 0)
				s->Devices = get_unsigned(key,value);
			else
				goto unknown;
			continue;

		case 'e':
			if (strcmp("exec",key) == 0)
				s->SysexecR = strtod(value,0);
			else
				goto unknown;
			continue;

		case 'f':
			if (strcmp("fork",key) == 0)
				s->SysforkR = strtod(value,0);
			else
				goto unknown;
			continue;

		case 'F':
			if (strcmp("Free",key) == 0)
				s->Pagesfree = get_unsigned(key,value);
			else if (strcmp("FsFree",key) == 0)
				s->FsFree = get_unsigned(key,value);
			else
				goto unknown;
			continue;

		case 'H':
			if (strcmp("Host",key) == 0) {
				if (strlen(value) > sizeof(s->Hostinfo))
					warn("hostname to long: %s\t%s\n",key,value);
				else
					strcpy(s->Hostinfo,value);
			} else if (strcmp("HErr",key) == 0)
				s->HardErrors = get_unsigned(key,value);
			else
				goto unknown;
			continue;

		case 'I':
			if (strcmp("Idle",key) == 0)
				s->IdleP = strtod(value,0);
			else if (strcmp("IpIErr",key) == 0)
				s->IpInErrs = get_unsigned(key,value);
			else if (strcmp("IpIDis",key) == 0)
				s->IpInDiscards = get_unsigned(key,value);
			else if (strcmp("IpODis",key) == 0)
				s->IpOutDiscards = get_unsigned(key,value);
			else if (strcmp("IpI",key) == 0)
				s->IpInR = strtod(value,0);
			else if (strcmp("IpO",key) == 0)
				s->IpOutR = strtod(value,0);
			else if (strcmp("Itv",key) == 0)
				;/* currently ignored */
			else
				goto unknown;
			continue;

		case 'i':
			if (strcmp("intr",key) == 0)
				s->IntrR = strtod(value,0);
			else if (strcmp("intrblk",key) == 0)
				s->IntrBlkR = strtod(value,0);
			else if (strcmp("inv_swtch",key) == 0)
				s->InvSwtchR = strtod(value,0);
			else
				goto unknown;
			continue;

		case 'L':
			if (strcmp("Load1",key) == 0) 
				s->loadavg[0] = strtod(value,0);
			else if (strcmp("Load5",key) == 0) 
				s->loadavg[1] = strtod(value,0);
			else if (strcmp("Load15",key) == 0) 
				s->loadavg[2] = strtod(value,0);
			else if (strcmp("Locked",key) == 0)
				s->Pageslocked = get_unsigned(key,value);
			else
				goto unknown;
			continue;

		case 'N':
			if (strcmp("NCPU",key) == 0)
				s->Ncpus = get_unsigned(key,value);
			else if (strcmp("NetI",key) == 0)
				s->NetInR = strtod(value,0);
			else if (strcmp("NetO",key) == 0)
				s->NetOutR = strtod(value,0);
			else if (strcmp("NfsI",key) == 0)
				s->NfsIR = strtod(value,0);
			else if (strcmp("NfsO",key) == 0)
				s->NfsOR = strtod(value,0);
			else if (strcmp("NfsD",key) == 0)
				s->NfsdR = strtod(value,0);
			else
				goto unknown;
			continue;
		
		case 'O':
			if (strcmp("OCPU",key) == 0)
				s->Ocpus = get_unsigned(key,value);
			else
				goto unknown;
			continue;
			
		case 'p':
			if (strcmp("pgswapin",key) == 0)
				s->PgswapinR = strtod(value,0);
			else if (strcmp("pgswapout",key) == 0)
				s->PgswapoutR = strtod(value,0);
			else if (strcmp("pswitch",key) == 0)
				s->PswitchR = strtod(value,0);
			else
				goto unknown;
			continue;
			
		case 'S':
			if (strcmp("SwapTotal",key) == 0)
				s->SwapTotal = get_unsigned(key,value);
			else if (strcmp("SwapFree",key) == 0)
				s->SwapFree = get_unsigned(key,value);
			else if (strcmp("SErr",key) == 0)
				s->SoftErrors = get_unsigned(key,value);
			else if (strcmp("Storage",key) == 0)
				s->Storage = get_unsigned(key,value);
			else
				goto unknown;
			continue;

		case 'u':
			if (strcmp("uid",key) == 0)
				parse_user(value,s);
			else
				goto unknown;
			continue;


		case 'Z':
			if (strcmp("Zpools",key) == 0)
				s->Zpools = get_unsigned(key,value);
			else if (strcmp("Zsize",key) == 0)
				s->Zsize = get_unsigned(key,value);
			else if (strcmp("Zfree",key) == 0)
				s->Zfree = get_unsigned(key,value);
			else
				goto unknown;
			continue;

		default:
			;
		}

		if (strcmp("TS",key) == 0)
			s->TS = get_unsigned(key,value);
		else if (strcmp("BT",key) == 0)
			s->BootTime = get_unsigned(key,value);
		else if (strcmp("User",key) == 0)
			s->UserP = strtod(value,0);
		else if (strcmp("Kernel",key) == 0)
			s->KernelP = strtod(value,0);
		else if (strcmp("xcall",key) == 0)
			s->XcallsR = strtod(value,0);
		else if (strcmp("lockf",key) == 0)
			s->LockfR = strtod(value,0);
		else if (strcmp("read",key) == 0)
			s->SysreadR = strtod(value,0);
		else if (strcmp("write",key) == 0)
			s->SyswriteR = strtod(value,0);
		else if (strcmp("Ram",key) == 0)
			s->Physmem = get_unsigned(key,value);
		else if (strcmp("scan",key) == 0)
			s->ScanR = strtod(value,0);
		else
unknown:
			info("ignoring unknown key: %s\n",key);
	}
	return s;
}

// vim:tw=0
